@@ -14,7 +14,8 @@
 # limitations under the License.
 
 import base64
-from collections import defaultdict
+import binascii
+from collections import defaultdict, OrderedDict
 from email.header import Header
 from hashlib import sha1, sha256, md5
 import hmac
@@ -24,14 +25,16 @@ import six
 from six.moves.urllib.parse import quote, unquote, parse_qsl
 import string
 
-from swift.common.utils import split_path
+from swift.common.utils import split_path, json, get_swift_info, \
+    close_if_possible
 from swift.common import swob
 from swift.common.http import HTTP_OK, HTTP_CREATED, HTTP_ACCEPTED, \
     HTTP_NO_CONTENT, HTTP_UNAUTHORIZED, HTTP_FORBIDDEN, HTTP_NOT_FOUND, \
     HTTP_CONFLICT, HTTP_UNPROCESSABLE_ENTITY, HTTP_REQUEST_ENTITY_TOO_LARGE, \
     HTTP_PARTIAL_CONTENT, HTTP_NOT_MODIFIED, HTTP_PRECONDITION_FAILED, \
     HTTP_REQUESTED_RANGE_NOT_SATISFIABLE, HTTP_LENGTH_REQUIRED, \
-    HTTP_BAD_REQUEST, HTTP_REQUEST_TIMEOUT, is_success
+    HTTP_BAD_REQUEST, HTTP_REQUEST_TIMEOUT, HTTP_SERVICE_UNAVAILABLE, \
+    is_success
 
 from swift.common.constraints import check_utf8
 from swift.proxy.controllers.base import get_container_info, \
@@ -44,13 +47,14 @@ from swift.common.middleware.s3api.controllers import ServiceController, \
     UploadController, UploadsController, VersioningController, \
     UnsupportedController, S3AclController, BucketController
 from swift.common.middleware.s3api.s3response import AccessDenied, \
-    InvalidArgument, InvalidDigest, \
+    InvalidArgument, InvalidDigest, BucketAlreadyOwnedByYou, \
     RequestTimeTooSkewed, S3Response, SignatureDoesNotMatch, \
     BucketAlreadyExists, BucketNotEmpty, EntityTooLarge, \
     InternalError, NoSuchBucket, NoSuchKey, PreconditionFailed, InvalidRange, \
     MissingContentLength, InvalidStorageClass, S3NotImplemented, InvalidURI, \
     MalformedXML, InvalidRequest, RequestTimeout, InvalidBucketName, \
-    BadDigest, AuthorizationHeaderMalformed, AuthorizationQueryParametersError
+    BadDigest, AuthorizationHeaderMalformed, \
+    AuthorizationQueryParametersError, ServiceUnavailable
 from swift.common.middleware.s3api.exception import NotS3Request, \
     BadSwiftRequest
 from swift.common.middleware.s3api.utils import utf8encode, \
@@ -89,6 +93,8 @@ def _header_strip(value):
         # behave as though it wasn't provided
         return None
     return stripped
+
+
 _header_strip.re = re.compile('^[\x00-\x20]*|[\x00-\x20]*$')
 
 
@@ -110,17 +116,47 @@ def _header_acl_property(resource):
                     doc='Get and set the %s acl property' % resource)
 
 
+class HashingInput(object):
+    """
+    wsgi.input wrapper to verify the hash of the input as it's read.
+    """
+    def __init__(self, reader, content_length, hasher, expected_hex_hash):
+        self._input = reader
+        self._to_read = content_length
+        self._hasher = hasher()
+        self._expected = expected_hex_hash
+
+    def read(self, size=None):
+        chunk = self._input.read(size)
+        self._hasher.update(chunk)
+        self._to_read -= len(chunk)
+        short_read = bool(chunk) if size is None else (len(chunk) < size)
+        if self._to_read < 0 or (short_read and self._to_read) or (
+                self._to_read == 0 and
+                self._hasher.hexdigest() != self._expected):
+            self.close()
+            # Since we don't return the last chunk, the PUT never completes
+            raise swob.HTTPUnprocessableEntity(
+                'The X-Amz-Content-SHA56 you specified did not match '
+                'what we received.')
+        return chunk
+
+    def close(self):
+        close_if_possible(self._input)
+
+
 class SigV4Mixin(object):
     """
     A request class mixin to provide S3 signature v4 functionality
     """
 
     def check_signature(self, secret):
+        secret = utf8encode(secret)
         user_signature = self.signature
-        derived_secret = 'AWS4' + secret
-        for scope_piece in self.scope:
+        derived_secret = b'AWS4' + secret
+        for scope_piece in self.scope.values():
             derived_secret = hmac.new(
-                derived_secret, scope_piece, sha256).digest()
+                derived_secret, scope_piece.encode('utf8'), sha256).digest()
         valid_signature = hmac.new(
             derived_secret, self.string_to_sign, sha256).hexdigest()
         return user_signature == valid_signature
@@ -176,6 +212,8 @@ class SigV4Mixin(object):
         err = None
         try:
             expires = int(self.params['X-Amz-Expires'])
+        except KeyError:
+            raise AccessDenied()
         except ValueError:
             err = 'X-Amz-Expires should be a number'
         else:
@@ -193,6 +231,15 @@ class SigV4Mixin(object):
         if int(self.timestamp) + expires < S3Timestamp.now():
             raise AccessDenied('Request has expired')
 
+    def _parse_credential(self, credential_string):
+        parts = credential_string.split("/")
+        # credential must be in following format:
+        # <access-key-id>/<date>/<AWS-region>/<AWS-service>/aws4_request
+        if not parts[0] or len(parts) != 5:
+            raise AccessDenied()
+        return dict(zip(['access', 'date', 'region', 'service', 'terminal'],
+                        parts))
+
     def _parse_query_authentication(self):
         """
         Parse v4 query authentication
@@ -205,10 +252,11 @@ class SigV4Mixin(object):
             raise InvalidArgument('X-Amz-Algorithm',
                                   self.params.get('X-Amz-Algorithm'))
         try:
-            cred_param = self.params['X-Amz-Credential'].split("/")
-            access = cred_param[0]
+            cred_param = self._parse_credential(
+                self.params['X-Amz-Credential'])
             sig = self.params['X-Amz-Signature']
-            expires = self.params['X-Amz-Expires']
+            if not sig:
+                raise AccessDenied()
         except KeyError:
             raise AccessDenied()
 
@@ -220,12 +268,26 @@ class SigV4Mixin(object):
 
         self._signed_headers = set(signed_headers.split(';'))
 
-        # credential must be in following format:
-        # <access-key-id>/<date>/<AWS-region>/<AWS-service>/aws4_request
-        if not all([access, sig, len(cred_param) == 5, expires]):
-            raise AccessDenied()
+        invalid_messages = {
+            'date': 'Invalid credential date "%s". This date is not the same '
+                    'as X-Amz-Date: "%s".',
+            'region': "Error parsing the X-Amz-Credential parameter; "
+                    "the region '%s' is wrong; expecting '%s'",
+            'service': 'Error parsing the X-Amz-Credential parameter; '
+                    'incorrect service "%s". This endpoint belongs to "%s".',
+            'terminal': 'Error parsing the X-Amz-Credential parameter; '
+                    'incorrect terminal "%s". This endpoint uses "%s".',
+        }
+        for key in ('date', 'region', 'service', 'terminal'):
+            if cred_param[key] != self.scope[key]:
+                kwargs = {}
+                if key == 'region':
+                    kwargs = {'region': self.scope['region']}
+                raise AuthorizationQueryParametersError(
+                    invalid_messages[key] % (cred_param[key], self.scope[key]),
+                    **kwargs)
 
-        return access, sig
+        return cred_param['access'], sig
 
     def _parse_header_authentication(self):
         """
@@ -237,30 +299,46 @@ class SigV4Mixin(object):
         """
 
         auth_str = self.headers['Authorization']
-        cred_param = auth_str.partition(
-            "Credential=")[2].split(',')[0].split("/")
-        access = cred_param[0]
+        cred_param = self._parse_credential(auth_str.partition(
+            "Credential=")[2].split(',')[0])
         sig = auth_str.partition("Signature=")[2].split(',')[0]
+        if not sig:
+            raise AccessDenied()
         signed_headers = auth_str.partition(
             "SignedHeaders=")[2].split(',', 1)[0]
-        # credential must be in following format:
-        # <access-key-id>/<date>/<AWS-region>/<AWS-service>/aws4_request
-        if not all([access, sig, len(cred_param) == 5]):
-            raise AccessDenied()
         if not signed_headers:
             # TODO: make sure if is it Malformed?
             raise AuthorizationHeaderMalformed()
 
+        invalid_messages = {
+            'date': 'Invalid credential date "%s". This date is not the same '
+                    'as X-Amz-Date: "%s".',
+            'region': "The authorization header is malformed; the region '%s' "
+                    "is wrong; expecting '%s'",
+            'service': 'The authorization header is malformed; incorrect '
+                    'service "%s". This endpoint belongs to "%s".',
+            'terminal': 'The authorization header is malformed; incorrect '
+                    'terminal "%s". This endpoint uses "%s".',
+        }
+        for key in ('date', 'region', 'service', 'terminal'):
+            if cred_param[key] != self.scope[key]:
+                kwargs = {}
+                if key == 'region':
+                    kwargs = {'region': self.scope['region']}
+                raise AuthorizationHeaderMalformed(
+                    invalid_messages[key] % (cred_param[key], self.scope[key]),
+                    **kwargs)
+
         self._signed_headers = set(signed_headers.split(';'))
 
-        return access, sig
+        return cred_param['access'], sig
 
     def _canonical_query_string(self):
         return '&'.join(
-            '%s=%s' % (quote(key, safe='-_.~'),
-                       quote(value, safe='-_.~'))
+            '%s=%s' % (swob.wsgi_quote(key, safe='-_.~'),
+                       swob.wsgi_quote(value, safe='-_.~'))
             for key, value in sorted(self.params.items())
-            if key not in ('Signature', 'X-Amz-Signature'))
+            if key not in ('Signature', 'X-Amz-Signature')).encode('ascii')
 
     def _headers_to_sign(self):
         """
@@ -309,7 +387,7 @@ class SigV4Mixin(object):
         """
         It won't require bucket name in canonical_uri for v4.
         """
-        return self.environ.get('RAW_PATH_INFO', self.path)
+        return swob.wsgi_to_bytes(self.environ.get('RAW_PATH_INFO', self.path))
 
     def _canonical_request(self):
         # prepare 'canonical_request'
@@ -327,7 +405,7 @@ class SigV4Mixin(object):
         #
 
         # 1. Add verb like: GET
-        cr = [self.method.upper()]
+        cr = [swob.wsgi_to_bytes(self.method.upper())]
 
         # 2. Add path like: /
         path = self._canonical_uri()
@@ -341,12 +419,12 @@ class SigV4Mixin(object):
         # host:iam.amazonaws.com
         # x-amz-date:20150830T123600Z
         headers_to_sign = self._headers_to_sign()
-        cr.append(''.join('%s:%s\n' % (key, value)
-                          for key, value in headers_to_sign))
+        cr.append(b''.join(swob.wsgi_to_bytes('%s:%s\n' % (key, value))
+                           for key, value in headers_to_sign))
 
         # 5. Add signed headers into canonical request like
         # content-type;host;x-amz-date
-        cr.append(';'.join(k for k, v in headers_to_sign))
+        cr.append(b';'.join(swob.wsgi_to_bytes(k) for k, v in headers_to_sign))
 
         # 6. Add payload string at the tail
         if 'X-Amz-Credential' in self.params:
@@ -358,22 +436,42 @@ class SigV4Mixin(object):
             raise InvalidRequest(msg)
         else:
             hashed_payload = self.headers['X-Amz-Content-SHA256']
-        cr.append(hashed_payload)
-        return '\n'.join(cr).encode('utf-8')
+            if hashed_payload != 'UNSIGNED-PAYLOAD':
+                if self.content_length == 0:
+                    if hashed_payload != sha256().hexdigest():
+                        raise BadDigest(
+                            'The X-Amz-Content-SHA56 you specified did not '
+                            'match what we received.')
+                elif self.content_length:
+                    self.environ['wsgi.input'] = HashingInput(
+                        self.environ['wsgi.input'],
+                        self.content_length,
+                        sha256,
+                        hashed_payload)
+                # else, length not provided -- Swift will kick out a
+                # 411 Length Required which will get translated back
+                # to a S3-style response in S3Request._swift_error_codes
+        cr.append(swob.wsgi_to_bytes(hashed_payload))
+        return b'\n'.join(cr)
 
     @property
     def scope(self):
-        return [self.timestamp.amz_date_format.split('T')[0],
-                self.location, SERVICE, 'aws4_request']
+        return OrderedDict([
+            ('date', self.timestamp.amz_date_format.split('T')[0]),
+            ('region', self.location),
+            ('service', SERVICE),
+            ('terminal', 'aws4_request'),
+        ])
 
     def _string_to_sign(self):
         """
         Create 'StringToSign' value in Amazon terminology for v4.
         """
-        return '\n'.join(['AWS4-HMAC-SHA256',
-                          self.timestamp.amz_date_format,
-                          '/'.join(self.scope),
-                          sha256(self._canonical_request()).hexdigest()])
+        return b'\n'.join([
+            b'AWS4-HMAC-SHA256',
+            self.timestamp.amz_date_format.encode('ascii'),
+            '/'.join(self.scope.values()).encode('utf8'),
+            sha256(self._canonical_request()).hexdigest().encode('ascii')])
 
     def signature_does_not_match_kwargs(self):
         kwargs = super(SigV4Mixin, self).signature_does_not_match_kwargs()
@@ -381,7 +479,7 @@ class SigV4Mixin(object):
         kwargs.update({
             'canonical_request': cr,
             'canonical_request_bytes': ' '.join(
-                format(ord(c), '02x') for c in cr),
+                format(ord(c), '02x') for c in cr.decode('latin1')),
         })
         return kwargs
 
@@ -414,8 +512,8 @@ class S3Request(swob.Request):
     bucket_acl = _header_acl_property('container')
     object_acl = _header_acl_property('object')
 
-    def __init__(self, env, app=None, slo_enabled=True,
-                 storage_domain='', location='US', force_request_log=False,
+    def __init__(self, env, app=None, slo_enabled=True, storage_domain='',
+                 location='us-east-1', force_request_log=False,
                  dns_compliant_bucket_names=True, allow_multipart_uploads=True,
                  allow_no_owner=False):
         # NOTE: app and allow_no_owner are not used by this class, need for
@@ -439,28 +537,21 @@ class S3Request(swob.Request):
             'string_to_sign': self.string_to_sign,
             'check_signature': self.check_signature,
         }
-        self.token = None
         self.account = None
         self.user_id = None
         self.slo_enabled = slo_enabled
 
-        # NOTE(andrey-mp): substitute authorization header for next modules
-        # in pipeline (s3token). it uses this and X-Auth-Token in specific
-        # format.
-        # (kota_): yeah, the reason we need this is s3token only supports
-        # v2 like header consists of AWS access:signature. Since the commit
-        # b626a3ca86e467fc7564eac236b9ee2efd49bdcc, the s3token is in swift3
-        # repo so probably we need to change s3token to support v4 format.
-        self.headers['Authorization'] = 'AWS %s:%s' % (
-            self.access_key, self.signature)
         # Avoids that swift.swob.Response replaces Location header value
         # by full URL when absolute path given. See swift.swob for more detail.
         self.environ['swift.leave_relative_location'] = True
 
     def check_signature(self, secret):
+        secret = utf8encode(secret)
         user_signature = self.signature
         valid_signature = base64.b64encode(hmac.new(
             secret, self.string_to_sign, sha1).digest()).strip()
+        if not six.PY2:
+            valid_signature = valid_signature.decode('ascii')
         return user_signature == valid_signature
 
     @property
@@ -529,7 +620,7 @@ class S3Request(swob.Request):
         return None
 
     def _parse_uri(self):
-        if not check_utf8(self.environ['PATH_INFO']):
+        if not check_utf8(swob.wsgi_to_str(self.environ['PATH_INFO'])):
             raise InvalidURI(self.path)
 
         if self.bucket_in_host:
@@ -587,8 +678,10 @@ class S3Request(swob.Request):
         :raises: NotS3Request
         """
         if self._is_query_auth:
+            self._validate_expire_param()
             return self._parse_query_authentication()
         elif self._is_header_auth:
+            self._validate_dates()
             return self._parse_header_authentication()
         else:
             # if this request is neither query auth nor header auth
@@ -603,7 +696,7 @@ class S3Request(swob.Request):
         # Expires header is a float since epoch
         try:
             ex = S3Timestamp(float(self.params['Expires']))
-        except ValueError:
+        except (KeyError, ValueError):
             raise AccessDenied()
 
         if S3Timestamp.now() > ex:
@@ -620,11 +713,6 @@ class S3Request(swob.Request):
         :raises: AccessDenied
         :raises: RequestTimeTooSkewed
         """
-        if self._is_query_auth:
-            self._validate_expire_param()
-            # TODO: make sure the case if timestamp param in query
-            return
-
         date_header = self.headers.get('Date')
         amz_date_header = self.headers.get('X-Amz-Date')
         if not date_header and not amz_date_header:
@@ -652,16 +740,16 @@ class S3Request(swob.Request):
                 raise InvalidArgument('Content-Length',
                                       self.environ['CONTENT_LENGTH'])
 
-        self._validate_dates()
-
         value = _header_strip(self.headers.get('Content-MD5'))
         if value is not None:
             if not re.match('^[A-Za-z0-9+/]+={0,2}$', value):
                 # Non-base64-alphabet characters in value.
                 raise InvalidDigest(content_md5=value)
             try:
-                self.headers['ETag'] = value.decode('base64').encode('hex')
-            except Exception:
+                self.headers['ETag'] = binascii.b2a_hex(
+                    binascii.a2b_base64(value))
+            except binascii.error:
+                # incorrect padding, most likely
                 raise InvalidDigest(content_md5=value)
 
             if len(self.headers['ETag']) != 32:
@@ -698,15 +786,27 @@ class S3Request(swob.Request):
         if 'x-amz-mfa' in self.headers:
             raise S3NotImplemented('MFA Delete is not supported.')
 
-        if 'x-amz-server-side-encryption' in self.headers:
-            raise S3NotImplemented('Server-side encryption is not supported.')
+        sse_value = self.headers.get('x-amz-server-side-encryption')
+        if sse_value is not None:
+            if sse_value not in ('aws:kms', 'AES256'):
+                raise InvalidArgument(
+                    'x-amz-server-side-encryption', sse_value,
+                    'The encryption method specified is not supported')
+            encryption_enabled = get_swift_info(admin=True)['admin'].get(
+                'encryption', {}).get('enabled')
+            if not encryption_enabled or sse_value != 'AES256':
+                raise S3NotImplemented(
+                    'Server-side encryption is not supported.')
 
         if 'x-amz-website-redirect-location' in self.headers:
             raise S3NotImplemented('Website redirection is not supported.')
 
         # https://docs.aws.amazon.com/AmazonS3/latest/API/sigv4-streaming.html
         # describes some of what would be required to support this
-        if 'aws-chunked' in self.headers.get('content-encoding', ''):
+        if any(['aws-chunked' in self.headers.get('content-encoding', ''),
+                'STREAMING-AWS4-HMAC-SHA256-PAYLOAD' == self.headers.get(
+                    'x-amz-content-sha256', ''),
+                'x-amz-decoded-content-length' in self.headers]):
             raise S3NotImplemented('Transfering payloads in multiple chunks '
                                    'using aws-chunked is not supported.')
 
@@ -734,10 +834,11 @@ class S3Request(swob.Request):
                                    'functionality that is not implemented',
                                    header='Transfer-Encoding')
 
-        if self.message_length() > max_length:
+        ml = self.message_length()
+        if ml and ml > max_length:
             raise MalformedXML()
 
-        if te or self.message_length():
+        if te or ml:
             # Limit the read similar to how SLO handles manifests
             body = self.body_file.read(max_length)
         else:
@@ -752,7 +853,7 @@ class S3Request(swob.Request):
             raise InvalidRequest('Missing required header for this request: '
                                  'Content-MD5')
 
-        digest = md5(body).digest().encode('base64').strip()
+        digest = base64.b64encode(md5(body).digest()).strip().decode('ascii')
         if self.environ['HTTP_CONTENT_MD5'] != digest:
             raise BadDigest(content_md5=self.environ['HTTP_CONTENT_MD5'])
 
@@ -836,9 +937,10 @@ class S3Request(swob.Request):
         """
         amz_headers = {}
 
-        buf = [self.method,
-               _header_strip(self.headers.get('Content-MD5')) or '',
-               _header_strip(self.headers.get('Content-Type')) or '']
+        buf = [swob.wsgi_to_bytes(wsgi_str) for wsgi_str in [
+            self.method,
+            _header_strip(self.headers.get('Content-MD5')) or '',
+            _header_strip(self.headers.get('Content-Type')) or '']]
 
         if 'headers_raw' in self.environ:  # eventlet >= 0.19.0
             # See https://github.com/eventlet/eventlet/commit/67ec999
@@ -857,18 +959,18 @@ class S3Request(swob.Request):
 
         if self._is_header_auth:
             if 'x-amz-date' in amz_headers:
-                buf.append('')
+                buf.append(b'')
             elif 'Date' in self.headers:
-                buf.append(self.headers['Date'])
+                buf.append(swob.wsgi_to_bytes(self.headers['Date']))
         elif self._is_query_auth:
-            buf.append(self.params['Expires'])
+            buf.append(swob.wsgi_to_bytes(self.params['Expires']))
         else:
             # Should have already raised NotS3Request in _parse_auth_info,
             # but as a sanity check...
             raise AccessDenied()
 
         for key, value in sorted(amz_headers.items()):
-            buf.append("%s:%s" % (key, value))
+            buf.append(swob.wsgi_to_bytes("%s:%s" % (key, value)))
 
         path = self._canonical_uri()
         if self.query_string:
@@ -880,10 +982,10 @@ class S3Request(swob.Request):
                 if key in ALLOWED_SUB_RESOURCES:
                     params.append('%s=%s' % (key, value) if value else key)
         if params:
-            buf.append('%s?%s' % (path, '&'.join(params)))
+            buf.append(swob.wsgi_to_bytes('%s?%s' % (path, '&'.join(params))))
         else:
-            buf.append(path)
-        return '\n'.join(buf)
+            buf.append(swob.wsgi_to_bytes(path))
+        return b'\n'.join(buf)
 
     def signature_does_not_match_kwargs(self):
         return {
@@ -891,7 +993,8 @@ class S3Request(swob.Request):
             'string_to_sign': self.string_to_sign,
             'signature_provided': self.signature,
             'string_to_sign_bytes': ' '.join(
-                format(ord(c), '02x') for c in self.string_to_sign),
+                format(ord(c), '02x')
+                for c in self.string_to_sign.decode('latin1')),
         }
 
     @property
@@ -1010,6 +1113,21 @@ class S3Request(swob.Request):
             env['HTTP_X_COPY_FROM'] = env['HTTP_X_AMZ_COPY_SOURCE']
             del env['HTTP_X_AMZ_COPY_SOURCE']
             env['CONTENT_LENGTH'] = '0'
+            if env.pop('HTTP_X_AMZ_METADATA_DIRECTIVE', None) == 'REPLACE':
+                env['HTTP_X_FRESH_METADATA'] = 'True'
+            else:
+                copy_exclude_headers = ('HTTP_CONTENT_DISPOSITION',
+                                        'HTTP_CONTENT_ENCODING',
+                                        'HTTP_CONTENT_LANGUAGE',
+                                        'CONTENT_TYPE',
+                                        'HTTP_EXPIRES',
+                                        'HTTP_CACHE_CONTROL',
+                                        'HTTP_X_ROBOTS_TAG')
+                for key in copy_exclude_headers:
+                    env.pop(key, None)
+                for key in list(env.keys()):
+                    if key.startswith('HTTP_X_OBJECT_META_'):
+                        del env[key]
 
         if self.force_request_log:
             env['swift.proxy_access_log_made'] = False
@@ -1017,8 +1135,6 @@ class S3Request(swob.Request):
         if method is not None:
             env['REQUEST_METHOD'] = method
 
-        env['HTTP_X_AUTH_TOKEN'] = self.token
-
         if obj:
             path = '/v1/%s/%s/%s' % (account, container, obj)
         elif container:
@@ -1087,6 +1203,7 @@ class S3Request(swob.Request):
                 ],
                 'PUT': [
                     HTTP_CREATED,
+                    HTTP_ACCEPTED,  # For SLO with heartbeating
                 ],
                 'POST': [
                     HTTP_ACCEPTED,
@@ -1099,6 +1216,20 @@ class S3Request(swob.Request):
 
         return code_map[method]
 
+    def _bucket_put_accepted_error(self, container, app):
+        sw_req = self.to_swift_req('HEAD', container, None)
+        info = get_container_info(sw_req.environ, app)
+        sysmeta = info.get('sysmeta', {})
+        try:
+            acl = json.loads(sysmeta.get('s3api-acl',
+                                         sysmeta.get('swift3-acl', '{}')))
+            owner = acl.get('Owner')
+        except (ValueError, TypeError, KeyError):
+            owner = None
+        if owner is None or owner == self.user_id:
+            raise BucketAlreadyOwnedByYou(container)
+        raise BucketAlreadyExists(container)
+
     def _swift_error_codes(self, method, container, obj, env, app):
         """
         Returns a dict from expected Swift error codes to the corresponding S3
@@ -1120,7 +1251,8 @@ class S3Request(swob.Request):
                     HTTP_NOT_FOUND: (NoSuchBucket, container),
                 },
                 'PUT': {
-                    HTTP_ACCEPTED: (BucketAlreadyExists, container),
+                    HTTP_ACCEPTED: (self._bucket_put_accepted_error, container,
+                                    app),
                 },
                 'POST': {
                     HTTP_NOT_FOUND: (NoSuchBucket, container),
@@ -1189,12 +1321,14 @@ class S3Request(swob.Request):
         sw_req = self.to_swift_req(method, container, obj, headers=headers,
                                    body=body, query=query)
 
-        sw_resp = sw_req.get_response(app)
-
-        # reuse account and tokens
-        _, self.account, _ = split_path(sw_resp.environ['PATH_INFO'],
-                                        2, 3, True)
-        self.account = utf8encode(self.account)
+        try:
+            sw_resp = sw_req.get_response(app)
+        except swob.HTTPException as err:
+            sw_resp = err
+        else:
+            # reuse account
+            _, self.account, _ = split_path(sw_resp.environ['PATH_INFO'],
+                                            2, 3, True)
 
         resp = S3Response.from_swift_resp(sw_resp)
         status = resp.status_int  # pylint: disable-msg=E1101
@@ -1202,10 +1336,11 @@ class S3Request(swob.Request):
         if not self.user_id:
             if 'HTTP_X_USER_NAME' in sw_resp.environ:
                 # keystone
-                self.user_id = \
-                    utf8encode("%s:%s" %
-                               (sw_resp.environ['HTTP_X_TENANT_NAME'],
-                                sw_resp.environ['HTTP_X_USER_NAME']))
+                self.user_id = "%s:%s" % (
+                    sw_resp.environ['HTTP_X_TENANT_NAME'],
+                    sw_resp.environ['HTTP_X_USER_NAME'])
+                if six.PY2 and not isinstance(self.user_id, bytes):
+                    self.user_id = self.user_id.encode('utf8')
             else:
                 # tempauth
                 self.user_id = self.access_key
@@ -1228,12 +1363,14 @@ class S3Request(swob.Request):
                 raise err_resp()
 
         if status == HTTP_BAD_REQUEST:
-            raise BadSwiftRequest(err_msg)
+            raise BadSwiftRequest(err_msg.decode('utf8'))
         if status == HTTP_UNAUTHORIZED:
             raise SignatureDoesNotMatch(
                 **self.signature_does_not_match_kwargs())
         if status == HTTP_FORBIDDEN:
             raise AccessDenied()
+        if status == HTTP_SERVICE_UNAVAILABLE:
+            raise ServiceUnavailable()
 
         raise InternalError('unexpected status code %d' % status)
 
@@ -1303,8 +1440,10 @@ class S3Request(swob.Request):
         else:
             # otherwise we do naive HEAD request with the authentication
             resp = self.get_response(app, 'HEAD', self.container_name, '')
+            headers = resp.sw_headers.copy()
+            headers.update(resp.sysmeta_headers)
             return headers_to_container_info(
-                resp.sw_headers, resp.status_int)  # pylint: disable-msg=E1101
+                headers, resp.status_int)  # pylint: disable-msg=E1101
 
     def gen_multipart_manifest_delete_query(self, app, obj=None):
         if not self.allow_multipart_uploads:
@@ -1323,8 +1462,8 @@ class S3AclRequest(S3Request):
     """
     S3Acl request object.
     """
-    def __init__(self, env, app, slo_enabled=True,
-                 storage_domain='', location='US', force_request_log=False,
+    def __init__(self, env, app, slo_enabled=True, storage_domain='',
+                 location='us-east-1', force_request_log=False,
                  dns_compliant_bucket_names=True, allow_multipart_uploads=True,
                  allow_no_owner=False):
         super(S3AclRequest, self).__init__(
@@ -1359,14 +1498,13 @@ class S3AclRequest(S3Request):
 
         _, self.account, _ = split_path(sw_resp.environ['PATH_INFO'],
                                         2, 3, True)
-        self.account = utf8encode(self.account)
 
         if 'HTTP_X_USER_NAME' in sw_resp.environ:
             # keystone
             self.user_id = "%s:%s" % (sw_resp.environ['HTTP_X_TENANT_NAME'],
                                       sw_resp.environ['HTTP_X_USER_NAME'])
-            self.user_id = utf8encode(self.user_id)
-            self.token = sw_resp.environ.get('HTTP_X_AUTH_TOKEN')
+            if six.PY2 and not isinstance(self.user_id, bytes):
+                self.user_id = self.user_id.encode('utf8')
         else:
             # tempauth
             self.user_id = self.access_key
@@ -1376,7 +1514,6 @@ class S3AclRequest(S3Request):
 
         # Need to skip S3 authorization on subsequent requests to prevent
         # overwriting the account in PATH_INFO
-        del self.headers['Authorization']
         del self.environ['s3api.auth_details']
 
     def to_swift_req(self, method, container, obj, query=None,
