@@ -13,15 +13,17 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+import base64
 import unittest
 from mock import patch, MagicMock
+import calendar
 from datetime import datetime
 import hashlib
 import mock
 import requests
 import json
-import copy
-from urllib import unquote, quote
+import six
+from six.moves.urllib.parse import unquote, quote
 
 import swift.common.middleware.s3api
 from swift.common.middleware.keystoneauth import KeystoneAuth
@@ -31,6 +33,7 @@ from swift.common.swob import Request
 from keystonemiddleware.auth_token import AuthProtocol
 from keystoneauth1.access import AccessInfoV2
 
+from test.unit import debug_logger
 from test.unit.common.middleware.s3api import S3ApiTestCase
 from test.unit.common.middleware.s3api.helpers import FakeSwift
 from test.unit.common.middleware.s3api.test_s3token import \
@@ -42,6 +45,54 @@ from swift.common.middleware.s3api.s3api import filter_factory, \
 from swift.common.middleware.s3api.s3token import S3Token
 
 
+class TestListingMiddleware(S3ApiTestCase):
+    def test_s3_etag_in_json(self):
+        # This translation happens all the time, even on normal swift requests
+        body_data = json.dumps([
+            {'name': 'obj1', 'hash': '0123456789abcdef0123456789abcdef'},
+            {'name': 'obj2', 'hash': 'swiftetag; s3_etag=mu-etag'},
+            {'name': 'obj2', 'hash': 'swiftetag; something=else'},
+            {'subdir': 'path/'},
+        ]).encode('ascii')
+        self.swift.register(
+            'GET', '/v1/a/c', swob.HTTPOk,
+            {'Content-Type': 'application/json; charset=UTF-8'},
+            body_data)
+
+        req = Request.blank('/v1/a/c')
+        status, headers, body = self.call_s3api(req)
+        self.assertEqual(json.loads(body), [
+            {'name': 'obj1', 'hash': '0123456789abcdef0123456789abcdef'},
+            {'name': 'obj2', 'hash': 'swiftetag', 's3_etag': '"mu-etag"'},
+            {'name': 'obj2', 'hash': 'swiftetag; something=else'},
+            {'subdir': 'path/'},
+        ])
+
+    def test_s3_etag_non_json(self):
+        self.swift.register(
+            'GET', '/v1/a/c', swob.HTTPOk,
+            {'Content-Type': 'application/json; charset=UTF-8'},
+            b'Not actually JSON')
+        req = Request.blank('/v1/a/c')
+        status, headers, body = self.call_s3api(req)
+        self.assertEqual(body, b'Not actually JSON')
+
+        # Yes JSON, but wrong content-type
+        body_data = json.dumps([
+            {'name': 'obj1', 'hash': '0123456789abcdef0123456789abcdef'},
+            {'name': 'obj2', 'hash': 'swiftetag; s3_etag=mu-etag'},
+            {'name': 'obj2', 'hash': 'swiftetag; something=else'},
+            {'subdir': 'path/'},
+        ]).encode('ascii')
+        self.swift.register(
+            'GET', '/v1/a/c', swob.HTTPOk,
+            {'Content-Type': 'text/plain; charset=UTF-8'},
+            body_data)
+        req = Request.blank('/v1/a/c')
+        status, headers, body = self.call_s3api(req)
+        self.assertEqual(body, body_data)
+
+
 class TestS3ApiMiddleware(S3ApiTestCase):
     def setUp(self):
         super(TestS3ApiMiddleware, self).setUp()
@@ -51,7 +102,7 @@ class TestS3ApiMiddleware(S3ApiTestCase):
     def test_non_s3_request_passthrough(self):
         req = Request.blank('/something')
         status, headers, body = self.call_s3api(req)
-        self.assertEqual(body, 'FAKE APP')
+        self.assertEqual(body, b'FAKE APP')
 
     def test_bad_format_authorization(self):
         req = Request.blank('/something',
@@ -115,7 +166,9 @@ class TestS3ApiMiddleware(S3ApiTestCase):
                 env[header] = value
 
             with patch('swift.common.middleware.s3api.s3request.'
-                       'S3Request._validate_headers'):
+                       'S3Request._validate_headers'), \
+                    patch('swift.common.middleware.s3api.s3request.'
+                          'S3Request._validate_dates'):
                 req = S3Request(env)
             return req.environ['s3api.auth_details']['string_to_sign']
 
@@ -216,8 +269,8 @@ class TestS3ApiMiddleware(S3ApiTestCase):
         req.content_type = 'text/plain'
         status, headers, body = self.call_s3api(req)
         self.assertEqual(status.split()[0], '200')
-        for _, _, headers in self.swift.calls_with_headers:
-            self.assertEqual(headers['Authorization'], 'AWS test:tester:X')
+        for _, path, headers in self.swift.calls_with_headers:
+            self.assertNotIn('Authorization', headers)
 
     def test_signed_urls_no_timestamp(self):
         expire = '2147483647'  # 19 Jan 2038 03:14:07
@@ -230,7 +283,7 @@ class TestS3ApiMiddleware(S3ApiTestCase):
         # for signed_url access and it also doesn't check timestamp
         self.assertEqual(status.split()[0], '200')
         for _, _, headers in self.swift.calls_with_headers:
-            self.assertEqual(headers['Authorization'], 'AWS test:tester:X')
+            self.assertNotIn('Authorization', headers)
 
     def test_signed_urls_invalid_expire(self):
         expire = 'invalid'
@@ -268,56 +321,97 @@ class TestS3ApiMiddleware(S3ApiTestCase):
         req = Request.blank(
             '/bucket/object'
             '?X-Amz-Algorithm=AWS4-HMAC-SHA256'
-            '&X-Amz-Credential=test:tester/20T20Z/US/s3/aws4_request'
+            '&X-Amz-Credential=test:tester/%s/us-east-1/s3/aws4_request'
             '&X-Amz-Date=%s'
             '&X-Amz-Expires=1000'
             '&X-Amz-SignedHeaders=host'
-            '&X-Amz-Signature=X' %
-            self.get_v4_amz_date_header(),
+            '&X-Amz-Signature=X' % (
+                self.get_v4_amz_date_header().split('T', 1)[0],
+                self.get_v4_amz_date_header()),
             headers={'Date': self.get_date_header()},
             environ={'REQUEST_METHOD': 'GET'})
         req.content_type = 'text/plain'
         status, headers, body = self.call_s3api(req)
         self.assertEqual(status.split()[0], '200', body)
         for _, _, headers in self.swift.calls_with_headers:
-            self.assertEqual('AWS test:tester:X', headers['Authorization'])
-            self.assertIn('X-Auth-Token', headers)
+            self.assertNotIn('Authorization', headers)
+            self.assertNotIn('X-Auth-Token', headers)
+
+    def test_signed_urls_v4_bad_credential(self):
+        def test(credential, message, extra=b''):
+            req = Request.blank(
+                '/bucket/object'
+                '?X-Amz-Algorithm=AWS4-HMAC-SHA256'
+                '&X-Amz-Credential=%s'
+                '&X-Amz-Date=%s'
+                '&X-Amz-Expires=1000'
+                '&X-Amz-SignedHeaders=host'
+                '&X-Amz-Signature=X' % (
+                    credential,
+                    self.get_v4_amz_date_header()),
+                headers={'Date': self.get_date_header()},
+                environ={'REQUEST_METHOD': 'GET'})
+            req.content_type = 'text/plain'
+            status, headers, body = self.call_s3api(req)
+            self.assertEqual(status.split()[0], '400', body)
+            self.assertEqual(self._get_error_code(body),
+                             'AuthorizationQueryParametersError')
+            self.assertEqual(self._get_error_message(body), message)
+            self.assertIn(extra, body)
+
+        dt = self.get_v4_amz_date_header().split('T', 1)[0]
+        test('test:tester/not-a-date/us-east-1/s3/aws4_request',
+             'Invalid credential date "not-a-date". This date is not the same '
+             'as X-Amz-Date: "%s".' % dt)
+        test('test:tester/%s/us-west-1/s3/aws4_request' % dt,
+             "Error parsing the X-Amz-Credential parameter; the region "
+             "'us-west-1' is wrong; expecting 'us-east-1'",
+             b'<Region>us-east-1</Region>')
+        test('test:tester/%s/us-east-1/not-s3/aws4_request' % dt,
+             'Error parsing the X-Amz-Credential parameter; incorrect service '
+             '"not-s3". This endpoint belongs to "s3".')
+        test('test:tester/%s/us-east-1/s3/not-aws4_request' % dt,
+             'Error parsing the X-Amz-Credential parameter; incorrect '
+             'terminal "not-aws4_request". This endpoint uses "aws4_request".')
 
     def test_signed_urls_v4_missing_x_amz_date(self):
-        req = Request.blank('/bucket/object'
-                            '?X-Amz-Algorithm=AWS4-HMAC-SHA256'
-                            '&X-Amz-Credential=test/20T20Z/US/s3/aws4_request'
-                            '&X-Amz-Expires=1000'
-                            '&X-Amz-SignedHeaders=host'
-                            '&X-Amz-Signature=X',
-                            environ={'REQUEST_METHOD': 'GET'})
+        req = Request.blank(
+            '/bucket/object'
+            '?X-Amz-Algorithm=AWS4-HMAC-SHA256'
+            '&X-Amz-Credential=test/20T20Z/us-east-1/s3/aws4_request'
+            '&X-Amz-Expires=1000'
+            '&X-Amz-SignedHeaders=host'
+            '&X-Amz-Signature=X',
+            environ={'REQUEST_METHOD': 'GET'})
         req.content_type = 'text/plain'
         status, headers, body = self.call_s3api(req)
         self.assertEqual(self._get_error_code(body), 'AccessDenied')
 
     def test_signed_urls_v4_invalid_algorithm(self):
-        req = Request.blank('/bucket/object'
-                            '?X-Amz-Algorithm=FAKE'
-                            '&X-Amz-Credential=test/20T20Z/US/s3/aws4_request'
-                            '&X-Amz-Date=%s'
-                            '&X-Amz-Expires=1000'
-                            '&X-Amz-SignedHeaders=host'
-                            '&X-Amz-Signature=X' %
-                            self.get_v4_amz_date_header(),
-                            environ={'REQUEST_METHOD': 'GET'})
+        req = Request.blank(
+            '/bucket/object'
+            '?X-Amz-Algorithm=FAKE'
+            '&X-Amz-Credential=test/20T20Z/us-east-1/s3/aws4_request'
+            '&X-Amz-Date=%s'
+            '&X-Amz-Expires=1000'
+            '&X-Amz-SignedHeaders=host'
+            '&X-Amz-Signature=X' %
+            self.get_v4_amz_date_header(),
+            environ={'REQUEST_METHOD': 'GET'})
         req.content_type = 'text/plain'
         status, headers, body = self.call_s3api(req)
         self.assertEqual(self._get_error_code(body), 'InvalidArgument')
 
     def test_signed_urls_v4_missing_signed_headers(self):
-        req = Request.blank('/bucket/object'
-                            '?X-Amz-Algorithm=AWS4-HMAC-SHA256'
-                            '&X-Amz-Credential=test/20T20Z/US/s3/aws4_request'
-                            '&X-Amz-Date=%s'
-                            '&X-Amz-Expires=1000'
-                            '&X-Amz-Signature=X' %
-                            self.get_v4_amz_date_header(),
-                            environ={'REQUEST_METHOD': 'GET'})
+        req = Request.blank(
+            '/bucket/object'
+            '?X-Amz-Algorithm=AWS4-HMAC-SHA256'
+            '&X-Amz-Credential=test/20T20Z/us-east-1/s3/aws4_request'
+            '&X-Amz-Date=%s'
+            '&X-Amz-Expires=1000'
+            '&X-Amz-Signature=X' %
+            self.get_v4_amz_date_header(),
+            environ={'REQUEST_METHOD': 'GET'})
         req.content_type = 'text/plain'
         status, headers, body = self.call_s3api(req)
         self.assertEqual(self._get_error_code(body),
@@ -338,14 +432,15 @@ class TestS3ApiMiddleware(S3ApiTestCase):
         self.assertEqual(self._get_error_code(body), 'AccessDenied')
 
     def test_signed_urls_v4_missing_signature(self):
-        req = Request.blank('/bucket/object'
-                            '?X-Amz-Algorithm=AWS4-HMAC-SHA256'
-                            '&X-Amz-Credential=test/20T20Z/US/s3/aws4_request'
-                            '&X-Amz-Date=%s'
-                            '&X-Amz-Expires=1000'
-                            '&X-Amz-SignedHeaders=host' %
-                            self.get_v4_amz_date_header(),
-                            environ={'REQUEST_METHOD': 'GET'})
+        req = Request.blank(
+            '/bucket/object'
+            '?X-Amz-Algorithm=AWS4-HMAC-SHA256'
+            '&X-Amz-Credential=test/20T20Z/us-east-1/s3/aws4_request'
+            '&X-Amz-Date=%s'
+            '&X-Amz-Expires=1000'
+            '&X-Amz-SignedHeaders=host' %
+            self.get_v4_amz_date_header(),
+            environ={'REQUEST_METHOD': 'GET'})
         req.content_type = 'text/plain'
         status, headers, body = self.call_s3api(req)
         self.assertEqual(self._get_error_code(body), 'AccessDenied')
@@ -390,9 +485,9 @@ class TestS3ApiMiddleware(S3ApiTestCase):
         self.assertEqual(req.environ['s3api.auth_details'], {
             'access_key': 'test:tester',
             'signature': 'hmac',
-            'string_to_sign': '\n'.join([
-                'PUT', '', '', date_header,
-                '/bucket/object?partNumber=1&uploadId=123456789abcdef']),
+            'string_to_sign': b'\n'.join([
+                b'PUT', b'', b'', date_header.encode('ascii'),
+                b'/bucket/object?partNumber=1&uploadId=123456789abcdef']),
             'check_signature': mock_cs})
 
     def test_invalid_uri(self):
@@ -413,8 +508,10 @@ class TestS3ApiMiddleware(S3ApiTestCase):
         self.assertEqual(self._get_error_code(body), 'InvalidDigest')
 
     def test_object_create_bad_md5_too_short(self):
-        too_short_digest = hashlib.md5('hey').hexdigest()[:-1]
-        md5_str = too_short_digest.encode('base64').strip()
+        too_short_digest = hashlib.md5(b'hey').digest()[:-1]
+        md5_str = base64.b64encode(too_short_digest).strip()
+        if not six.PY2:
+            md5_str = md5_str.decode('ascii')
         req = Request.blank(
             '/bucket/object',
             environ={'REQUEST_METHOD': 'PUT',
@@ -425,8 +522,10 @@ class TestS3ApiMiddleware(S3ApiTestCase):
         self.assertEqual(self._get_error_code(body), 'InvalidDigest')
 
     def test_object_create_bad_md5_too_long(self):
-        too_long_digest = hashlib.md5('hey').hexdigest() + 'suffix'
-        md5_str = too_long_digest.encode('base64').strip()
+        too_long_digest = hashlib.md5(b'hey').digest() + b'suffix'
+        md5_str = base64.b64encode(too_long_digest).strip()
+        if not six.PY2:
+            md5_str = md5_str.decode('ascii')
         req = Request.blank(
             '/bucket/object',
             environ={'REQUEST_METHOD': 'PUT',
@@ -455,6 +554,15 @@ class TestS3ApiMiddleware(S3ApiTestCase):
         status, headers, body = self.call_s3api(req)
         self.assertEqual(self._get_error_code(body), 'InvalidStorageClass')
 
+    def test_invalid_ssc(self):
+        req = Request.blank('/',
+                            environ={'REQUEST_METHOD': 'GET',
+                                     'HTTP_AUTHORIZATION': 'AWS X:Y:Z'},
+                            headers={'x-amz-server-side-encryption': 'invalid',
+                                     'Date': self.get_date_header()})
+        status, headers, body = self.call_s3api(req)
+        self.assertEqual(self._get_error_code(body), 'InvalidArgument')
+
     def _test_unsupported_header(self, header, value=None):
         if value is None:
             value = 'value'
@@ -469,8 +577,27 @@ class TestS3ApiMiddleware(S3ApiTestCase):
     def test_mfa(self):
         self._test_unsupported_header('x-amz-mfa')
 
-    def test_server_side_encryption(self):
-        self._test_unsupported_header('x-amz-server-side-encryption')
+    @mock.patch.object(utils, '_swift_admin_info', new_callable=dict)
+    def test_server_side_encryption(self, mock_info):
+        sse_header = 'x-amz-server-side-encryption'
+        self._test_unsupported_header(sse_header, 'AES256')
+        self._test_unsupported_header(sse_header, 'aws:kms')
+        utils.register_swift_info('encryption', admin=True, enabled=False)
+        self._test_unsupported_header(sse_header, 'AES256')
+        self._test_unsupported_header(sse_header, 'aws:kms')
+        utils.register_swift_info('encryption', admin=True, enabled=True)
+        # AES256 now works
+        self.swift.register('PUT', '/v1/AUTH_X/bucket/object',
+                            swob.HTTPCreated, {}, None)
+        req = Request.blank('/bucket/object',
+                            environ={'REQUEST_METHOD': 'PUT',
+                                     'HTTP_AUTHORIZATION': 'AWS X:Y:Z'},
+                            headers={sse_header: 'AES256',
+                                     'Date': self.get_date_header()})
+        status, headers, body = self.call_s3api(req)
+        self.assertEqual(status, '200 OK')
+        # ...but aws:kms continues to fail
+        self._test_unsupported_header(sse_header, 'aws:kms')
 
     def test_website_redirect_location(self):
         self._test_unsupported_header('x-amz-website-redirect-location')
@@ -486,6 +613,11 @@ class TestS3ApiMiddleware(S3ApiTestCase):
         # > That is, you can specify your custom content-encoding when using
         # > Signature Version 4 streaming API.
         self._test_unsupported_header('Content-Encoding', 'aws-chunked,gzip')
+        # Some clients skip the content-encoding,
+        # such as minio-go and aws-sdk-java
+        self._test_unsupported_header('x-amz-content-sha256',
+                                      'STREAMING-AWS4-HMAC-SHA256-PAYLOAD')
+        self._test_unsupported_header('x-amz-decoded-content-length')
 
     def test_object_tagging(self):
         self._test_unsupported_header('x-amz-tagging')
@@ -579,13 +711,13 @@ class TestS3ApiMiddleware(S3ApiTestCase):
             with self.assertRaises(ValueError) as cm:
                 self.s3api.check_pipeline(self.conf)
             self.assertIn('expected auth between s3api and proxy-server',
-                          cm.exception.message)
+                          cm.exception.args[0])
 
             pipeline.return_value = 'proxy-server'
             with self.assertRaises(ValueError) as cm:
                 self.s3api.check_pipeline(self.conf)
             self.assertIn("missing filters ['s3api']",
-                          cm.exception.message)
+                          cm.exception.args[0])
 
     def test_s3api_initialization_with_disabled_pipeline_check(self):
         with patch("swift.common.middleware.s3api.s3api.loadcontext"), \
@@ -619,12 +751,14 @@ class TestS3ApiMiddleware(S3ApiTestCase):
     def test_signature_v4(self):
         environ = {
             'REQUEST_METHOD': 'GET'}
+        authz_header = 'AWS4-HMAC-SHA256 ' + ', '.join([
+            'Credential=test:tester/%s/us-east-1/s3/aws4_request' %
+            self.get_v4_amz_date_header().split('T', 1)[0],
+            'SignedHeaders=host;x-amz-date',
+            'Signature=X',
+        ])
         headers = {
-            'Authorization':
-                'AWS4-HMAC-SHA256 '
-                'Credential=test:tester/20130524/US/s3/aws4_request, '
-                'SignedHeaders=host;x-amz-date,'
-                'Signature=X',
+            'Authorization': authz_header,
             'X-Amz-Date': self.get_v4_amz_date_header(),
             'X-Amz-Content-SHA256': '0123456789'}
         req = Request.blank('/bucket/object', environ=environ, headers=headers)
@@ -632,8 +766,8 @@ class TestS3ApiMiddleware(S3ApiTestCase):
         status, headers, body = self.call_s3api(req)
         self.assertEqual(status.split()[0], '200', body)
         for _, _, headers in self.swift.calls_with_headers:
-            self.assertEqual('AWS test:tester:X', headers['Authorization'])
-            self.assertIn('X-Auth-Token', headers)
+            self.assertEqual(authz_header, headers['Authorization'])
+            self.assertNotIn('X-Auth-Token', headers)
 
     def test_signature_v4_no_date(self):
         environ = {
@@ -641,7 +775,7 @@ class TestS3ApiMiddleware(S3ApiTestCase):
         headers = {
             'Authorization':
                 'AWS4-HMAC-SHA256 '
-                'Credential=test:tester/20130524/US/s3/aws4_request, '
+                'Credential=test:tester/20130524/us-east-1/s3/aws4_request, '
                 'SignedHeaders=host;range;x-amz-date,'
                 'Signature=X',
             'X-Amz-Content-SHA256': '0123456789'}
@@ -657,9 +791,9 @@ class TestS3ApiMiddleware(S3ApiTestCase):
         headers = {
             'Authorization':
                 'AWS4-HMAC-SHA256 '
-                'Credential=test:tester/20130524/US/s3/aws4_request, '
+                'Credential=test:tester/%s/us-east-1/s3/aws4_request, '
                 'SignedHeaders=host;x-amz-date,'
-                'Signature=X',
+                'Signature=X' % self.get_v4_amz_date_header().split('T', 1)[0],
             'X-Amz-Date': self.get_v4_amz_date_header()}
         req = Request.blank('/bucket/object', environ=environ, headers=headers)
         req.content_type = 'text/plain'
@@ -671,7 +805,7 @@ class TestS3ApiMiddleware(S3ApiTestCase):
             'Missing required header for this request: x-amz-content-sha256')
 
     def test_signature_v4_bad_authorization_string(self):
-        def test(auth_str, error, msg):
+        def test(auth_str, error, msg, extra=b''):
             environ = {
                 'REQUEST_METHOD': 'GET'}
             headers = {
@@ -684,23 +818,52 @@ class TestS3ApiMiddleware(S3ApiTestCase):
             status, headers, body = self.call_s3api(req)
             self.assertEqual(self._get_error_code(body), error)
             self.assertEqual(self._get_error_message(body), msg)
+            self.assertIn(extra, body)
 
         auth_str = ('AWS4-HMAC-SHA256 '
                     'SignedHeaders=host;x-amz-date,'
                     'Signature=X')
         test(auth_str, 'AccessDenied', 'Access Denied.')
 
-        auth_str = ('AWS4-HMAC-SHA256 '
-                    'Credential=test:tester/20130524/US/s3/aws4_request, '
-                    'Signature=X')
+        auth_str = (
+            'AWS4-HMAC-SHA256 '
+            'Credential=test:tester/20130524/us-east-1/s3/aws4_request, '
+            'Signature=X')
         test(auth_str, 'AuthorizationHeaderMalformed',
              'The authorization header is malformed; the authorization '
              'header requires three components: Credential, SignedHeaders, '
              'and Signature.')
 
         auth_str = ('AWS4-HMAC-SHA256 '
-                    'Credential=test:tester/20130524/US/s3/aws4_request, '
-                    'SignedHeaders=host;x-amz-date')
+                    'Credential=test:tester/%s/us-west-2/s3/aws4_request, '
+                    'Signature=X, SignedHeaders=host;x-amz-date' %
+                    self.get_v4_amz_date_header().split('T', 1)[0])
+        test(auth_str, 'AuthorizationHeaderMalformed',
+             "The authorization header is malformed; "
+             "the region 'us-west-2' is wrong; expecting 'us-east-1'",
+             b'<Region>us-east-1</Region>')
+
+        auth_str = ('AWS4-HMAC-SHA256 '
+                    'Credential=test:tester/%s/us-east-1/not-s3/aws4_request, '
+                    'Signature=X, SignedHeaders=host;x-amz-date' %
+                    self.get_v4_amz_date_header().split('T', 1)[0])
+        test(auth_str, 'AuthorizationHeaderMalformed',
+             'The authorization header is malformed; '
+             'incorrect service "not-s3". This endpoint belongs to "s3".')
+
+        auth_str = ('AWS4-HMAC-SHA256 '
+                    'Credential=test:tester/%s/us-east-1/s3/not-aws4_request, '
+                    'Signature=X, SignedHeaders=host;x-amz-date' %
+                    self.get_v4_amz_date_header().split('T', 1)[0])
+        test(auth_str, 'AuthorizationHeaderMalformed',
+             'The authorization header is malformed; '
+             'incorrect terminal "not-aws4_request". '
+             'This endpoint uses "aws4_request".')
+
+        auth_str = (
+            'AWS4-HMAC-SHA256 '
+            'Credential=test:tester/20130524/us-east-1/s3/aws4_request, '
+            'SignedHeaders=host;x-amz-date')
         test(auth_str, 'AccessDenied', 'Access Denied.')
 
     def test_canonical_string_v4(self):
@@ -720,13 +883,16 @@ class TestS3ApiMiddleware(S3ApiTestCase):
                     '27ae41e4649b934ca495991b7852b855',
                 'HTTP_AUTHORIZATION':
                     'AWS4-HMAC-SHA256 '
-                    'Credential=X:Y/dt/reg/host/blah, '
+                    'Credential=X:Y/20110909/us-east-1/s3/aws4_request, '
                     'SignedHeaders=content-md5;content-type;date, '
                     'Signature=x',
             }
+            fake_time = calendar.timegm((2011, 9, 9, 23, 36, 0))
             env.update(environ)
             with patch('swift.common.middleware.s3api.s3request.'
-                       'S3Request._validate_headers'):
+                       'S3Request._validate_headers'), \
+                    patch('swift.common.middleware.s3api.utils.time.time',
+                          return_value=fake_time):
                 req = SigV4Request(env, location=self.conf.location)
             return req
 
@@ -741,8 +907,8 @@ class TestS3ApiMiddleware(S3ApiTestCase):
                     patch.object(swift.common.middleware.s3api.s3request,
                                  'SERVICE', 'host'):
                 req = _get_req(path, environ)
-                hash_in_sts = req._string_to_sign().split('\n')[3]
-                self.assertEqual(hash_val, hash_in_sts)
+                hash_in_sts = req._string_to_sign().split(b'\n')[3]
+                self.assertEqual(hash_val, hash_in_sts.decode('ascii'))
                 self.assertTrue(req.check_signature(
                     'wJalrXUtnFEMI/K7MDENG+bPxRfiCYEXAMPLEKEY'))
 
@@ -803,7 +969,7 @@ class TestS3ApiMiddleware(S3ApiTestCase):
                    'validate_bucket_name'):
             verify('27ba31df5dbc6e063d8f87d62eb07143'
                    'f7f271c5330a917840586ac1c85b6f6b',
-                   unquote('/%E1%88%B4'), env)
+                   swob.wsgi_unquote('/%E1%88%B4'), env)
 
         # get-vanilla-query-order-key
         env = {
@@ -889,7 +1055,7 @@ class TestS3ApiMiddleware(S3ApiTestCase):
         headers = {
             'Authorization':
                 'AWS4-HMAC-SHA256 '
-                'Credential=test/20130524/US/s3/aws4_request_A, '
+                'Credential=test/20130524/us-east-1/s3/aws4_request_A, '
                 'SignedHeaders=hostA;rangeA;x-amz-dateA,'
                 'Signature=X',
             'X-Amz-Date': self.get_v4_amz_date_header(),
@@ -897,13 +1063,14 @@ class TestS3ApiMiddleware(S3ApiTestCase):
 
         # and then, different auth info (Credential, SignedHeaders, Signature)
         # in query
-        req = Request.blank('/bucket/object'
-                            '?X-Amz-Algorithm=AWS4-HMAC-SHA256'
-                            '&X-Amz-Credential=test/20T20Z/US/s3/aws4_requestB'
-                            '&X-Amz-SignedHeaders=hostB'
-                            '&X-Amz-Signature=Y',
-                            environ={'REQUEST_METHOD': 'GET'},
-                            headers=headers)
+        req = Request.blank(
+            '/bucket/object'
+            '?X-Amz-Algorithm=AWS4-HMAC-SHA256'
+            '&X-Amz-Credential=test/20T20Z/us-east-1/s3/aws4_requestB'
+            '&X-Amz-SignedHeaders=hostB'
+            '&X-Amz-Signature=Y',
+            environ={'REQUEST_METHOD': 'GET'},
+            headers=headers)
         req.content_type = 'text/plain'
         status, headers, body = self.call_s3api(req)
         # FIXME: should this failed as 400 or pass via query auth?
@@ -911,12 +1078,13 @@ class TestS3ApiMiddleware(S3ApiTestCase):
         self.assertEqual(status.split()[0], '403', body)
 
         # But if we are missing Signature in query param
-        req = Request.blank('/bucket/object'
-                            '?X-Amz-Algorithm=AWS4-HMAC-SHA256'
-                            '&X-Amz-Credential=test/20T20Z/US/s3/aws4_requestB'
-                            '&X-Amz-SignedHeaders=hostB',
-                            environ={'REQUEST_METHOD': 'GET'},
-                            headers=headers)
+        req = Request.blank(
+            '/bucket/object'
+            '?X-Amz-Algorithm=AWS4-HMAC-SHA256'
+            '&X-Amz-Credential=test/20T20Z/us-east-1/s3/aws4_requestB'
+            '&X-Amz-SignedHeaders=hostB',
+            environ={'REQUEST_METHOD': 'GET'},
+            headers=headers)
         req.content_type = 'text/plain'
         status, headers, body = self.call_s3api(req)
         self.assertEqual(status.split()[0], '403', body)
@@ -928,6 +1096,7 @@ class TestS3ApiMiddleware(S3ApiTestCase):
         self.s3_token = S3Token(
             self.keystone_auth, {'auth_uri': 'https://fakehost/identity'})
         self.s3api = S3ApiMiddleware(self.s3_token, self.conf)
+        self.s3api.logger = debug_logger()
         req = Request.blank(
             '/bucket',
             environ={'REQUEST_METHOD': 'PUT'},
@@ -939,12 +1108,12 @@ class TestS3ApiMiddleware(S3ApiTestCase):
                             swob.HTTPOk, {}, None)
         with patch.object(self.s3_token, '_json_request') as mock_req:
             mock_resp = requests.Response()
-            mock_resp._content = json.dumps(GOOD_RESPONSE_V2)
+            mock_resp._content = json.dumps(GOOD_RESPONSE_V2).encode('ascii')
             mock_resp.status_code = 201
             mock_req.return_value = mock_resp
 
             status, headers, body = self.call_s3api(req)
-            self.assertEqual(body, '')
+            self.assertEqual(body, b'')
             self.assertEqual(1, mock_req.call_count)
 
     def test_s3api_with_only_s3_token_v3(self):
@@ -954,6 +1123,7 @@ class TestS3ApiMiddleware(S3ApiTestCase):
         self.s3_token = S3Token(
             self.keystone_auth, {'auth_uri': 'https://fakehost/identity'})
         self.s3api = S3ApiMiddleware(self.s3_token, self.conf)
+        self.s3api.logger = debug_logger()
         req = Request.blank(
             '/bucket',
             environ={'REQUEST_METHOD': 'PUT'},
@@ -965,12 +1135,12 @@ class TestS3ApiMiddleware(S3ApiTestCase):
                             swob.HTTPOk, {}, None)
         with patch.object(self.s3_token, '_json_request') as mock_req:
             mock_resp = requests.Response()
-            mock_resp._content = json.dumps(GOOD_RESPONSE_V3)
+            mock_resp._content = json.dumps(GOOD_RESPONSE_V3).encode('ascii')
             mock_resp.status_code = 200
             mock_req.return_value = mock_resp
 
             status, headers, body = self.call_s3api(req)
-            self.assertEqual(body, '')
+            self.assertEqual(body, b'')
             self.assertEqual(1, mock_req.call_count)
 
     def test_s3api_with_s3_token_and_auth_token(self):
@@ -982,6 +1152,7 @@ class TestS3ApiMiddleware(S3ApiTestCase):
         self.s3_token = S3Token(
             self.auth_token, {'auth_uri': 'https://fakehost/identity'})
         self.s3api = S3ApiMiddleware(self.s3_token, self.conf)
+        self.s3api.logger = debug_logger()
         req = Request.blank(
             '/bucket',
             environ={'REQUEST_METHOD': 'PUT'},
@@ -994,8 +1165,11 @@ class TestS3ApiMiddleware(S3ApiTestCase):
         with patch.object(self.s3_token, '_json_request') as mock_req:
             with patch.object(self.auth_token,
                               '_do_fetch_token') as mock_fetch:
+                # sanity check
+                self.assertIn('id', GOOD_RESPONSE_V2['access']['token'])
                 mock_resp = requests.Response()
-                mock_resp._content = json.dumps(GOOD_RESPONSE_V2)
+                mock_resp._content = json.dumps(
+                    GOOD_RESPONSE_V2).encode('ascii')
                 mock_resp.status_code = 201
                 mock_req.return_value = mock_resp
 
@@ -1005,21 +1179,28 @@ class TestS3ApiMiddleware(S3ApiTestCase):
                 mock_fetch.return_value = (MagicMock(), mock_access_info)
 
                 status, headers, body = self.call_s3api(req)
-                self.assertEqual(body, '')
+                # Even though s3token got a token back from keystone, we drop
+                # it on the floor, resulting in a 401 Unauthorized at
+                # `swift.common.middleware.keystoneauth` because
+                # keystonemiddleware's auth_token strips out all auth headers,
+                # significantly 'X-Identity-Status'. Without a token, it then
+                # sets 'X-Identity-Status: Invalid' and never contacts
+                # Keystone.
+                self.assertEqual('403 Forbidden', status)
                 self.assertEqual(1, mock_req.call_count)
-                # With X-Auth-Token, auth_token will call _do_fetch_token to
-                # connect to keystone in auth_token, again
-                self.assertEqual(1, mock_fetch.call_count)
+                # it never even tries to contact keystone
+                self.assertEqual(0, mock_fetch.call_count)
 
-    def test_s3api_with_s3_token_no_pass_token_to_auth_token(self):
+    def test_s3api_with_only_s3_token_in_s3acl(self):
         self.swift = FakeSwift()
         self.keystone_auth = KeystoneAuth(
             self.swift, {'operator_roles': 'swift-user'})
-        self.auth_token = AuthProtocol(
-            self.keystone_auth, {'delay_auth_decision': 'True'})
         self.s3_token = S3Token(
-            self.auth_token, {'auth_uri': 'https://fakehost/identity'})
+            self.keystone_auth, {'auth_uri': 'https://fakehost/identity'})
+
+        self.conf['s3_acl'] = True
         self.s3api = S3ApiMiddleware(self.s3_token, self.conf)
+        self.s3api.logger = debug_logger()
         req = Request.blank(
             '/bucket',
             environ={'REQUEST_METHOD': 'PUT'},
@@ -1027,40 +1208,21 @@ class TestS3ApiMiddleware(S3ApiTestCase):
                      'Date': self.get_date_header()})
         self.swift.register('PUT', '/v1/AUTH_TENANT_ID/bucket',
                             swob.HTTPCreated, {}, None)
-        self.swift.register('HEAD', '/v1/AUTH_TENANT_ID',
-                            swob.HTTPOk, {}, None)
+        # For now, s3 acl commits the bucket owner acl via POST
+        # after PUT container so we need to register the resposne here
+        self.swift.register('POST', '/v1/AUTH_TENANT_ID/bucket',
+                            swob.HTTPNoContent, {}, None)
+        self.swift.register('TEST', '/v1/AUTH_TENANT_ID',
+                            swob.HTTPMethodNotAllowed, {}, None)
         with patch.object(self.s3_token, '_json_request') as mock_req:
-            with patch.object(self.auth_token,
-                              '_do_fetch_token') as mock_fetch:
-                mock_resp = requests.Response()
-                no_token_id_good_resp = copy.deepcopy(GOOD_RESPONSE_V2)
-                # delete token id
-                del no_token_id_good_resp['access']['token']['id']
-                mock_resp._content = json.dumps(no_token_id_good_resp)
-                mock_resp.status_code = 201
-                mock_req.return_value = mock_resp
+            mock_resp = requests.Response()
+            mock_resp._content = json.dumps(GOOD_RESPONSE_V2).encode('ascii')
+            mock_resp.status_code = 201
+            mock_req.return_value = mock_resp
 
-                mock_access_info = AccessInfoV2(GOOD_RESPONSE_V2)
-                mock_access_info.will_expire_soon = \
-                    lambda stale_duration: False
-                mock_fetch.return_value = (MagicMock(), mock_access_info)
-
-                status, headers, body = self.call_s3api(req)
-                # No token provided from keystone result in 401 Unauthorized
-                # at `swift.common.middleware.keystoneauth` because auth_token
-                # will remove all auth headers including 'X-Identity-Status'[1]
-                # and then, set X-Identity-Status: Invalid at [2]
-                #
-                # 1: https://github.com/openstack/keystonemiddleware/blob/
-                #    master/keystonemiddleware/auth_token/__init__.py#L620
-                # 2: https://github.com/openstack/keystonemiddleware/blob/
-                #    master/keystonemiddleware/auth_token/__init__.py#L627-L629
-
-                self.assertEqual('403 Forbidden', status)
-                self.assertEqual(1, mock_req.call_count)
-                # if no token provided from keystone, we can skip the call to
-                # fetch the token
-                self.assertEqual(0, mock_fetch.call_count)
+            status, headers, body = self.call_s3api(req)
+            self.assertEqual(body, b'')
+            self.assertEqual(1, mock_req.call_count)
 
 if __name__ == '__main__':
     unittest.main()
