@@ -26,6 +26,7 @@ import random
 
 from collections import defaultdict
 from datetime import datetime
+import six
 from six.moves import urllib
 from swift.container import reconciler
 from swift.container.server import gen_resp_headers
@@ -105,8 +106,10 @@ class FakeInternalClient(reconciler.InternalClient):
             else:
                 timestamp, content_type = timestamp, 'application/x-put'
             storage_policy_index, path = item
+            if six.PY2 and isinstance(path, six.text_type):
+                path = path.encode('utf-8')
             account, container_name, obj_name = split_path(
-                path.encode('utf-8'), 0, 3, rest_with_last=True)
+                path, 0, 3, rest_with_last=True)
             self.accounts[account][container_name].append(
                 (obj_name, storage_policy_index, timestamp, content_type))
         for account_name, containers in self.accounts.items():
@@ -124,7 +127,8 @@ class FakeInternalClient(reconciler.InternalClient):
                     if storage_policy_index is None and not obj_name:
                         # empty container
                         continue
-                    obj_path = container_path + '/' + obj_name
+                    obj_path = swob.str_to_wsgi(
+                        container_path + '/' + obj_name)
                     ts = Timestamp(timestamp)
                     headers = {'X-Timestamp': ts.normal,
                                'X-Backend-Timestamp': ts.internal}
@@ -139,27 +143,32 @@ class FakeInternalClient(reconciler.InternalClient):
                     # strings, so normalize here
                     if isinstance(timestamp, numbers.Number):
                         timestamp = '%f' % timestamp
+                    if six.PY2:
+                        obj_name = obj_name.decode('utf-8')
+                        timestamp = timestamp.decode('utf-8')
                     obj_data = {
                         'bytes': 0,
                         # listing data is unicode
-                        'name': obj_name.decode('utf-8'),
+                        'name': obj_name,
                         'last_modified': last_modified,
-                        'hash': timestamp.decode('utf-8'),
+                        'hash': timestamp,
                         'content_type': content_type,
                     }
                     container_listing_data.append(obj_data)
                 container_listing_data.sort(key=operator.itemgetter('name'))
                 # register container listing response
                 container_headers = {}
-                container_qry_string = '?format=json&marker=&end_marker='
+                container_qry_string = \
+                    '?format=json&marker=&end_marker=&prefix='
                 self.app.register('GET', container_path + container_qry_string,
                                   swob.HTTPOk, container_headers,
                                   json.dumps(container_listing_data))
                 if container_listing_data:
                     obj_name = container_listing_data[-1]['name']
                     # client should quote and encode marker
-                    end_qry_string = '?format=json&marker=%s&end_marker=' % (
-                        urllib.parse.quote(obj_name.encode('utf-8')))
+                    end_qry_string = \
+                        '?format=json&marker=%s&end_marker=&prefix=' % (
+                            urllib.parse.quote(obj_name.encode('utf-8')))
                     self.app.register('GET', container_path + end_qry_string,
                                       swob.HTTPOk, container_headers,
                                       json.dumps([]))
@@ -171,11 +180,11 @@ class FakeInternalClient(reconciler.InternalClient):
             # register account response
             account_listing_data.sort(key=operator.itemgetter('name'))
             account_headers = {}
-            account_qry_string = '?format=json&marker=&end_marker='
+            account_qry_string = '?format=json&marker=&end_marker=&prefix='
             self.app.register('GET', account_path + account_qry_string,
                               swob.HTTPOk, account_headers,
                               json.dumps(account_listing_data))
-            end_qry_string = '?format=json&marker=%s&end_marker=' % (
+            end_qry_string = '?format=json&marker=%s&end_marker=&prefix=' % (
                 urllib.parse.quote(account_listing_data[-1]['name']))
             self.app.register('GET', account_path + end_qry_string,
                               swob.HTTPOk, account_headers,
@@ -362,12 +371,12 @@ class TestReconcilerUtils(unittest.TestCase):
             direct_head.side_effect = stub_resp_headers
             oldest_spi = reconciler.direct_get_container_policy_index(
                 self.fake_ring, 'a', 'con')
-        self.assertEqual(oldest_spi, None)
+        self.assertIsNone(oldest_spi)
 
     def test_get_container_policy_index_for_deleted(self):
         mock_path = 'swift.container.reconciler.direct_head_container'
         headers = container_resp_headers(
-            status_changed_at=Timestamp(time.time()).internal,
+            status_changed_at=Timestamp.now().internal,
             storage_policy_index=1,
         )
         stub_resp_headers = [
@@ -551,7 +560,7 @@ class TestReconcilerUtils(unittest.TestCase):
                 oldest_spi = reconciler.direct_get_container_policy_index(
                     self.fake_ring, 'a', 'con')
         # expired
-        self.assertEqual(oldest_spi, None)
+        self.assertIsNone(oldest_spi)
 
     def test_direct_delete_container_entry(self):
         mock_path = 'swift.common.direct_client.http_connect'
@@ -564,7 +573,7 @@ class TestReconcilerUtils(unittest.TestCase):
                 'partition': partition, 'method': method, 'path': path,
                 'headers': headers, 'query_string': query_string})
 
-        x_timestamp = Timestamp(time.time())
+        x_timestamp = Timestamp.now()
         headers = {'x-timestamp': x_timestamp.internal}
         fake_hc = fake_http_connect(200, 200, 200, give_connect=test_connect)
         with mock.patch(mock_path, fake_hc):
@@ -597,7 +606,7 @@ class TestReconcilerUtils(unittest.TestCase):
                 mock.patch('eventlet.greenpool.DEBUG', False):
             rv = reconciler.direct_delete_container_entry(
                 self.fake_ring, 'a', 'c', 'o')
-        self.assertEqual(rv, None)
+        self.assertIsNone(rv)
         self.assertEqual(len(mock_direct_delete.mock_calls), 3)
 
     def test_add_to_reconciler_queue(self):
@@ -704,7 +713,7 @@ class TestReconcilerUtils(unittest.TestCase):
 
 
 def listing_qs(marker):
-    return "?format=json&marker=%s&end_marker=" % \
+    return "?format=json&marker=%s&end_marker=&prefix=" % \
         urllib.parse.quote(marker.encode('utf-8'))
 
 
@@ -750,7 +759,7 @@ class TestReconciler(unittest.TestCase):
         with mock.patch.multiple(reconciler, **items) as mocks:
             self.mock_delete_container_entry = \
                 mocks['direct_delete_container_entry']
-            with mock.patch('time.time', mock_time_iter.next):
+            with mock.patch('time.time', lambda: next(mock_time_iter)):
                 self.reconciler.run_once()
 
         return [c[1][1:4] for c in
@@ -972,7 +981,10 @@ class TestReconciler(unittest.TestCase):
         # functions where we call them with (account, container, obj)
         obj_name = u"AUTH_bob/c \u062a/o1 \u062a"
         # anytime we talk about a call made to swift for a path
-        obj_path = obj_name.encode('utf-8')
+        if six.PY2:
+            obj_path = obj_name.encode('utf-8')
+        else:
+            obj_path = obj_name.encode('utf-8').decode('latin-1')
         # this mock expects unquoted unicode because it handles container
         # listings as well as paths
         self._mock_listing({
@@ -1159,7 +1171,7 @@ class TestReconciler(unittest.TestCase):
         q_path = '.misplaced_objects/%s' % container
         self._mock_listing({
             (None, "/%s/1:/AUTH_bob/c/o1" % q_path): q_ts,
-            (1, '/AUTH_bob/c/o1'): q_ts - 0.00001,  # slightly older
+            (1, '/AUTH_bob/c/o1'): q_ts - 1,  # slightly older
         })
         self._mock_oldest_spi({'c': 0})
         deleted_container_entries = self._run_once()
@@ -1191,7 +1203,7 @@ class TestReconciler(unittest.TestCase):
 
     def test_src_object_unavailable_with_slightly_newer_tombstone(self):
         # should be some sort of retry case
-        q_ts = float(Timestamp(time.time()))
+        q_ts = float(Timestamp.now())
         container = str(int(q_ts // 3600 * 3600))
         q_path = '.misplaced_objects/%s' % container
         self._mock_listing({
@@ -1230,7 +1242,7 @@ class TestReconciler(unittest.TestCase):
 
     def test_src_object_unavailable_server_error(self):
         # should be some sort of retry case
-        q_ts = float(Timestamp(time.time()))
+        q_ts = float(Timestamp.now())
         container = str(int(q_ts // 3600 * 3600))
         q_path = '.misplaced_objects/%s' % container
         self._mock_listing({
@@ -1399,7 +1411,7 @@ class TestReconciler(unittest.TestCase):
         self._mock_listing({
             (None, "/.misplaced_objects/3600/1:/AUTH_bob/c/o1"): 3679.2019,
             (1, "/AUTH_bob/c/o1"): 3679.2019,
-            (0, "/AUTH_bob/c/o1"): 3679.2019 + 0.00001,  # slightly newer
+            (0, "/AUTH_bob/c/o1"): 3679.2019 + 1,  # slightly newer
         })
         self._mock_oldest_spi({'c': 0})
         deleted_container_entries = self._run_once()
@@ -1440,7 +1452,7 @@ class TestReconciler(unittest.TestCase):
         self._mock_listing({
             (None, "/.misplaced_objects/36000/1:/AUTH_bob/c/o1"): 36123.38393,
             (1, "/AUTH_bob/c/o1"): 36123.38393,
-            (0, "/AUTH_bob/c/o1"): 36123.38393 - 0.00001,  # slightly older
+            (0, "/AUTH_bob/c/o1"): 36123.38393 - 1,  # slightly older
         })
         self._mock_oldest_spi({'c': 0})
         deleted_container_entries = self._run_once()
@@ -1583,7 +1595,7 @@ class TestReconciler(unittest.TestCase):
         self.assertEqual(self.reconciler.stats['retry'], 1)
 
     def test_object_move_no_such_object_no_tombstone_recent(self):
-        q_ts = float(Timestamp(time.time()))
+        q_ts = float(Timestamp.now())
         container = str(int(q_ts // 3600 * 3600))
         q_path = '.misplaced_objects/%s' % container
 
@@ -1615,7 +1627,7 @@ class TestReconciler(unittest.TestCase):
         self.assertEqual(deleted_container_entries, [])
 
     def test_object_move_no_such_object_no_tombstone_ancient(self):
-        queue_ts = float(Timestamp(time.time())) - \
+        queue_ts = float(Timestamp.now()) - \
             self.reconciler.reclaim_age * 1.1
         container = str(int(queue_ts // 3600 * 3600))
 
